package cmsc420_f20;

import java.util.ArrayList;

/**
 * An implementation of a extended binary search tree.
 *
 * Key-value pairs are stored at the leaves and keys are stored in the internal
 * nodes.
 *
 * Generic Elements: The tree is parameterized by two types, Key and Value. The
 * first is used for searching and the second is simply stored.
 */

public class XBSTree<Key extends Comparable<Key>, Value> {

	private final boolean DEBUG = false; // produce extra debugging output

	// -----------------------------------------------------------------
	// Node processing - Used internally only
	//
	// There are two types of nodes, external and internal. All nodes
	// key values. External nodes also store the associated value.
	// Internal nodes also store left and right child pointers.
	// -----------------------------------------------------------------

	/**
	 * A node of the tree. This is an abstract object, which is extended for
	 * internal and external nodes.
	 */
	private abstract class Node { // generic node type
		Key key; // the key

		Node(Key key) { // basic constructor
			this.key = key;
		}

		/**
		 * Key comparisons. The "keyGoesOnLeft" is used to determine whether a key goes
		 * in the left child of an internal node. The "keyEquals" is used to identify
		 * matching keys for insertion (duplicate key), find, and deletion.
		 *
		 * @param x The key.
		 */
		boolean keyGoesOnLeft(Key x) {
			return x.compareTo(key) < 0;
		}

		boolean keyMatches(Key x) {
			return x.compareTo(key) == 0;
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		abstract Value find(Key x); // find key in subtree

		abstract Node insert(Key x, Value v) throws Exception; // insert key-value into subtree

		abstract Node delete(Key x) throws Exception; // delete key from subtree

		abstract ArrayList<String> getPreorderList(); // list entries in preorder

		// -----------------------------------------------------------------
		// Additional operations
		// -----------------------------------------------------------------

		abstract Value getMin(); // get the minimum key in subtree

		abstract Value getMax(); // get the maximum key in subtree

		abstract Value findUp(Key x); // find next that is greater or equal

		abstract Value findDown(Key x); // find next that is lesser or equal

		abstract String debugPrint(String prefix); // print for debugging

		abstract public String toString(); // for debugging
	}

	// -----------------------------------------------------------------
	// Internal node
	// -----------------------------------------------------------------

	/**
	 * An internal node of the tree. The associated key is just used for splitting
	 * left and right subtrees.
	 */
	private class InternalNode extends Node {
		Node left; // children
		Node right;

		/**
		 * Basic constructor.
		 */
		InternalNode(Key key, Node left, Node right) {
			super(key);
			this.left = left;
			this.right = right;
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Find key in internal node.
		 */
		Value find(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				return left.find(x);
			} else { // x goes in right subtree
				return right.find(x);
			}
		}

		/**
		 * Insert into this subtree.
		 *
		 * This just passes the request to the appropriate child.
		 */
		Node insert(Key x, Value v) throws Exception {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				left = left.insert(x, v);
			} else { // x goes in right subtree
				right = right.insert(x, v);
			}
			return this;
		}

		/**
		 * Delete from this subtree.
		 *
		 * Special processing is needed when the child deleted is an external
		 * node. In this case, the deletion function returns null, and we
		 * return a reference to the opposite child, which effectively deletes
		 * this node.
		 */
		Node delete(Key x) throws Exception {
			if (keyGoesOnLeft(x)) { // delete from left
				left = left.delete(x);
				if (left == null)
					return right; // subtree is gone, return sibling
				else
					return this;
			} else { // delete from right
				right = right.delete(x);
				if (right == null)
					return left; // subtree is gone, return sibling
				else
					return this;
			}
		}

		/**
		 * Get preorder list.
		 *
		 * Adds current key in parentheses, followed by left and right
		 * subtrees.
		 */
		ArrayList<String> getPreorderList() {
			ArrayList<String> list = new ArrayList<String>();
			list.add("(" + key.toString() + ")"); // add this node
			list.addAll(left.getPreorderList()); // add left
			list.addAll(right.getPreorderList()); // add right
			return list;
		}

		// -----------------------------------------------------------------
		// Additional operations
		// -----------------------------------------------------------------

		/**
		 * Get the value associated with the minimum key.
		 *
		 */
		Value getMin() {
			return left.getMin();
		}

		/**
		 * Get the value associated with the maximum key.
		 *
		 */
		Value getMax() {
			return right.getMax();
		}

		/**
		 * Find next value with key greater or equal.
		 */
		Value findUp(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				Value v = left.findUp(x);
				if (v == null)
					return right.getMin();
				else
					return v;
			} else { // x goes in right subtree
				return right.findUp(x);
			}
		}

		/**
		 * Find next value with key lesser or equal.
		 */
		Value findDown(Key x) {
			if (keyGoesOnLeft(x)) { // x goes in left subtree
				return left.findDown(x);
			} else { // x goes in right subtree
				Value v = right.findDown(x);
				if (v == null)
					return left.getMax();
				else
					return v;
			}
		}

		/**
		 * Debug print subtree.
		 *
		 * This prints the subtree for debugging. I prefer to print in
		 * reverse inorder, since the tree's structure is evident by
		 * turning the printed picture 90 degrees clockwise.
		 */
		String debugPrint(String prefix) {
			return right.debugPrint(prefix + "| ") + System.lineSeparator() + prefix + toString()
					+ System.lineSeparator() + left.debugPrint(prefix + "| ");
		}

		/**
		 * Print node contents (for debugging).
		 */
		public String toString() {
			return "(" + key.toString() + ")";
		}
	}

	// -----------------------------------------------------------------
	// External node
	// -----------------------------------------------------------------

	/**
	 * An external node of the tree. This stores adds a value.
	 */
	private class ExternalNode extends Node {

		final Value value; // the associated value object

		/**
		 * Constructor from a value.
		 */
		ExternalNode(Key key, Value value) {
			super(key);
			this.value = value;
		}

		// -----------------------------------------------------------------
		// Standard dictionary operations
		// -----------------------------------------------------------------

		/**
		 * Find key in external node.
		 */
		Value find(Key x) {
			if (keyMatches(x))
				return value;
			else
				return null;
		}

		/**
		 * Insertion at an external node.
		 *
		 * There is a subtlety here. When x is smaller than the key, we
		 * create an internal node storing key, with x on the left and this
		 * node on the right. For this to happen, x must be smaller than
		 * any key in tree, so the new internal node's key does not match
		 * any other key in the tree.
		 */
		Node insert(Key x, Value v) throws Exception {
			if (keyMatches(x)) {
				throw new Exception("Insertion of duplicate key");
			} else {
				ExternalNode q = new ExternalNode(x, v); // new external node
				if (keyGoesOnLeft(x)) { // insert on left side (x must be smallest overall)
					return new InternalNode(key, q, this);
				} else { // insert on right side (more typical)
					return new InternalNode(x, this, q);
				}
			}
		}

		/**
		 * Delete key from this node.
		 *
		 * We expect to find the deleted key here. If not, we throw an
		 * exception. Otherwise, we remove the node by returning null (see
		 * InternalNode deletion for more information).
		 */
		Node delete(Key x) throws Exception {
			if (keyMatches(x)) { // found it
				return null;
			} else {
				throw new Exception("Deletion of nonexistent key");
			}
		}

		/**
		 * Get preorder list. This just returns an encoding of the data in this node.
		 */
		ArrayList<String> getPreorderList() {
			ArrayList<String> list = new ArrayList<String>();
			list.add("[" + key.toString() + " " + value.toString() + "]"); // add this node
			return list;
		}

		// -----------------------------------------------------------------
		// Additional operations
		// -----------------------------------------------------------------

		/**
		 * Get the value associated with the minimum key.
		 */
		Value getMin() {
			return value;
		}

		/**
		 * Get the value associated with the maximum key.
		 */
		Value getMax() {
			return value;
		}

		/**
		 * Find next value with key greater or equal.
		 */
		Value findUp(Key x) {
			if (x.compareTo(key) <= 0)
				return value;
			else
				return null;
		}

		/**
		 * Find next value with key lesser or equal.
		 */
		Value findDown(Key x) {
			if (x.compareTo(key) >= 0)
				return value;
			else
				return null;
		}

		/**
		 * Debug print subtree.
		 */
		String debugPrint(String prefix) {
			return prefix + toString();
		}

		/**
		 * Convert to string (for debugging).
		 */
		public String toString() {
			return "[" + value.toString() + "]";
		}
	}

	// -----------------------------------------------------------------
	// Private member data
	// -----------------------------------------------------------------

	private Node root; // root of the tree

	// -----------------------------------------------------------------
	// Public members
	// -----------------------------------------------------------------

	/**
	 * Creates an empty tree.
	 */
	public XBSTree() {
		root = null;
	}

	// -----------------------------------------------------------------
	// Standard dictionary operations
	// -----------------------------------------------------------------

	/**
	 * Find value associated with a given key.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value find(Key x) {
		if (root == null)
			return null;
		else
			return root.find(x);
	}

	/**
	 * Insert a key-value pair.
	 *
	 * @param x The key
	 * @param v The value
	 * @throws Exception If they key already appears in the dictionary
	 */
	public void insert(Key x, Value v) throws Exception {
		if (DEBUG) {
			System.out.println("Tree: Inserting " + x);
		}
		if (root == null) { // insert the key
			root = new ExternalNode(x, v);
		} else {
			root = root.insert(x, v);
		}
		if (DEBUG) {
			System.out.println("Tree: After inserting " + x + System.lineSeparator() + debugPrint("  "));
		}
	}

	/**
	 * Delete a key.
	 *
	 * @param x The key to be deleted
	 * @throws Exception If they key already appears in the dictionary
	 */
	public void delete(Key x) throws Exception {
		if (DEBUG) {
			System.out.println("Tree: Deleting " + x);
		}
		if (root == null) { // empty tree?
			throw new Exception("Deletion of nonexistent key");
		} else {
			root = root.delete(x); // delete the key
		}
		if (DEBUG) {
			System.out.println("Tree: After deleting " + x + System.lineSeparator() + debugPrint("  "));
		}
	}

	/**
	 * Get a list of entries in preorder
	 *
	 * @return ArrayList of string encoded items in preorder
	 */
	public ArrayList<String> getPreorderList() {
		if (root == null) { // empty tree?
			return new ArrayList<String>(); // return an empty list
		} else {
			return root.getPreorderList();
		}
	}

	// -----------------------------------------------------------------
	// Additional operations
	// -----------------------------------------------------------------

	/**
	 * Get value associated with minimum key.
	 *
	 * @return A reference to the associated value, or null if tree empty
	 */
	public Value getMin() {
		if (root == null)
			return null;
		else
			return root.getMin();
	}

	/**
	 * Get value associated with maximum key.
	 *
	 * @return A reference to the associated value, or null if tree empty
	 */
	public Value getMax() {
		if (root == null)
			return null;
		else
			return root.getMax();
	}

	/**
	 * Find next value with key greater or equal.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value findUp(Key x) {
		if (root == null)
			return null;
		else
			return root.findUp(x);
	}

	/**
	 * Find next value with key lesser or equal.
	 *
	 * @param x The key being sought
	 * @return A reference to the associated value
	 */
	public Value findDown(Key x) {
		if (root == null)
			return null;
		else
			return root.findDown(x);
	}

	/**
	 * Clear the tree, removing all entries.
	 *
	 */
	public void clear() {
		root = null;
		if (DEBUG) {
			System.out.println("Tree: After clear " + System.lineSeparator() + debugPrint("  "));
		}
	}

	// -----------------------------------------------------------------
	// Operations that are not part of the assignment
	// -----------------------------------------------------------------

	/**
	 * Print the tree for debugging purposes
	 * 
	 * @param prefix A string prefix that precedes each line.
	 * @return A string representing the tree
	 */
	String debugPrint(String prefix) {
		if (root != null)
			return root.debugPrint(prefix);
		else
			return new String();
	}

}
